/* DEBUG4MC.PRG
   by Doug Schwarz

   A routine to determine values for Manifest Constants (#defines)
      in a given PRG (and, optionally, any #included files).
   Ultimately intended to be incorporated into the Clipper debugger.
      For now, run from the debugger's expression window,
      passing the current PRG name (with or without extension)
      as a parameter, e.g.:

                  ?mc('thisprg')
                  ?mc('thisprg.prg')

   Required from NH_Funcs:  Window()        Greater()
                            DSCenter()      LSI()

   Things known to confuse this routine include:

      #defined values based on formulas
      #defined non-character values containing spaces
      #defines which are >1 line long

      For example (from Clipper's BOX.CH header file):

        #define B_SINGLE    ( CHR(218) + CHR(196) + CHR(191) + CHR(179) + ;
                              CHR(217) + CHR(196) + CHR(192) + CHR(179)   )

   At present this routine does not parse such #defines correctly, and
   therefore does not show the proper values for these constants.
*/

#include 'inkey.ch'
#include 'achoice.ch'

FUNCTION MC(prgfile)

local prgstr, prglin,;
      aDef  := {},;
      aDisp := {},;
      commode, oldscrn,;
      oldcolor := setcolor()
local dx,deflin,mcname,mctype,mcvalue,namelen,aWid,;
      gnamlen := 0,;
      gvallen := 0,;
      inclen  := 0
private incinc,;
        dcc,icc,scc,pcc,dcr,icr,scr,pcr,;
        aPrg := {},;
        defcnt := inccnt := sortcnt := prgcnt := 0

BEGIN SEQUENCE

/* Get PRG file to analyze, and make sure we can find it
   (Eventually want to extract file name from Debugger environment,
    and use Debugger path info when locating the file)                */

if prgfile = NIL
  alert( 'No filename passed' )
  break
endif

prgfile := alltrim(upper(prgfile))
if empty(prgfile)
  alert( 'No filename passed' )
  break
else
  if right(prgfile,4) # ".PRG"
    prgfile += ".PRG"
  endif
  if !file(prgfile)
    alert( prgfile + ' not found' )
    break
  endif
endif
* row += 2

window('cc',8,30,177,0,setcolor(),@oldscrn)
@ c1+2,dscenter(c2,c4,"                    " + space(3)) say "PRG lines analyzed: "
pcr := row()
pcc := col()
@ c1+3,dscenter(c2,c4,"                    " + space(3)) say "#DEFINEs found:     "
dcr := row()
dcc := col()
@ c1+4,dscenter(c2,c4,"PRG lines analyzed: " + space(3)) say "#INCLUDEs found:    "
icr := row()
icc := col()
@ c1+5,dscenter(c2,c4,"                    " + space(3)) say "Parsing #DEFINEs:   "
scr := row()
scc := col()

ANALYPRG(prgfile)

/* aPrg now contains all functional #defines.  Each element of aPrg is a
      two-element array containing: { Line of code containing #define,
                                      Name of file from which code comes }
   Each line starts w/ "#define" and there are no K_ENTERs remaining
   We assume each "#define" is followed by 1+ spaces or tabs,
      and then a constant name
   The name MAY be followed by 1+ spaces or tabs, which then MAY be
      followed by a value
   The value MAY be followed by 1+ spaces or tabs, which then MAY be
      followed by a comment, indicated by the presence
      of && or // after the value. */

if incinc = NIL // no #includes found
  incinc := .F.
endif

if len(aPrg) > 0
for dx = 1 to len(aPrg)

  @ scr,scc say lsi(++sortcnt)

  deflin := trim(aPrg[dx][1]) // strip trailing spaces
  deflin := strtran(deflin,chr(K_TAB)," ") // replace all tabs w/ spaces

  if "&&" $ deflin // strip comment if any
    deflin := substr(deflin,1,at("&&",deflin)-1)
  elseif "//" $ deflin
    deflin := substr(deflin,1,at("//",deflin)-1)
  endif
  deflin := trim(deflin) // strip trailing spaces (value <=> comment)

  deflin := substr(deflin,8) // strip leading "#define"
  deflin := ltrim(deflin)    // strip leading spaces (#define <=> name)

  BEGIN SEQUENCE

  if " " $ deflin // spaces remaining (name <=> value)

    namelen := at(" ",deflin) - 1
    gnamlen := greater(gnamlen,namelen)

    mcname  := left(deflin,namelen)
    if "(" $ mcname // line #defines a function, not a constant
      break
    endif

    deflin  := substr(deflin,namelen+1) // strip name
    deflin  := ltrim(deflin) // strip leading spaces (name <=> value)

    DO CASE
    case left(deflin,1) = '"' .or. left(deflin,1) = "'" .or. ;
                                   left(deflin,1) = "["  // type C

        mctype  := "C"
        mcvalue := substr(deflin,2,len(deflin)-2) // strip delimiters
        gvallen := greater(gvallen,len(mcvalue))

    case left(upper(deflin),4) = 'CTOD' // type D

        mctype  := "D"
        mcvalue := substr(deflin,7,8)
        gvallen := greater(gvallen,8)

    case left(deflin,1) = "." .and. substr(deflin,3,1) = "." // type L

        mctype  := "L"
        mcvalue := left(deflin,3)
        gvallen := greater(gvallen,3)

    case left(deflin,1) $ "-1234567890." // type N

        mctype  := "N"
        if at(" ",deflin) = 0
          mcvalue := deflin
        else
          mcvalue := left( deflin, at(" ",deflin)-1 )
        endif
        gvallen := greater(gvallen,len(mcvalue))

    ENDCASE

  else // no spaces after name, so no value
    mcname  := deflin
    namelen := len(mcname)
    gnamlen := greater(gnamlen,namelen)
    mctype  := NIL
    mcvalue := NIL
  endif

  if ascan( aDef, { |def| def[1] == mcname } ) = 0 // mcname not previously defined
    aadd( aDef, {mcname,mctype,mcvalue,aPrg[dx][2]} )
    inclen := greater( len(aPrg[dx][2]), inclen )
  endif

  END SEQUENCE

next  // #define
endif // #defines exist

/* aDef now contains 1 element per each constant.
   Each element of aDef is a 4-element array containing: { Constant Name,;
                                                           Type,;
                                                           Assigned Value,;
                                                           Source File }
   All elements are stored in the array as strings for display purposes. */

@ c1+6,dscenter(c2,c4,"                    " + space(3)) say "Sorting #defines ..."
aDef := asort( aDef, NIL, NIL, { |x,y| x[1] < y[1] } )

@ c1+6,dscenter(c2,c4,"                    " + space(3)) say "Compiling all info ..."
aeval( aDef, { |e| aadd( aDisp, padr(e[1],gnamlen) + "" + ;
                                if( e[2] = NIL,;
                                    "NIL" + space(gvallen-1),;
                                    e[2] + "" + padl(e[3],gvallen) ) + ;
                                if( incinc,;
                                    "" + padr(e[4],inclen),;
                                    "" ) ) } )

restscreen(c1,c2,c3,c4,oldscrn)
window('br',12,gnamlen+gvallen+inclen+5,-1,0,setcolor(),@oldscrn)
achoice(c1+1,c2+1,c3-1,c4-1,aDisp,NIL,'mcafunc')

setcolor(oldcolor)
restscreen(c1,c2,c3,c4,oldscrn)

END SEQUENCE

return

/*-----------------------------------------------------------*/

FUNCTION ANALYPRG(prgfile)
/* Analyze the file PRGFILE, extracting all #defines and adding them to aPrg.
   If an #include is found (and user so chooses),
      call self recursively, passing the name of the include file.
   Returns NIL (but adds to the private aPrg) */

local prgstr,commode,prglin,delim1,delim2,incstart,incstop,incfile,;
      incstr,istr,px,;
      incpath := ""
static aIPath  := {}

prgstr := memoread(prgfile) // read contents of prg file into string
commode := .F.

do while at( chr(K_ENTER), prgstr ) > 0

  @ pcr,pcc say lsi(++prgcnt)
  prglin := upper(alltrim(substr( prgstr, 1, at(chr(K_ENTER),prgstr)-1 )))

  if commode // block comment in process

    if right(prglin,2) == "*/" // end block comment
      commode := .F.
    endif

  else // not in comment

    if left(prglin,2) = "/*" // begin block comment

      /* Note: This assumes block comments do not begin or end in the middle
         of a line!  That is, we're assuming that when block comments begins
         the /* characters are the first non-blank characters in the line,
         and when it ends the terminating characters are the last non-blank
         characters in the line. */

      if !right(prglin,2) = "*/" // block comment does NOT end on same line
        commode := .T.
      endif

    elseif left(prglin,7) == "#DEFINE"

      @ dcr,dcc say lsi(++defcnt)
      aadd( aPrg, {prglin,prgfile} )

    elseif left(prglin,8) == "#INCLUDE"

      if incinc = NIL
        if alert('Include #INCLUDEs?',{"No","Yes"}) = 2
          incinc := .T.
        else
          incinc := .f.
        endif
      endif

      if incinc

        BEGIN SEQUENCE

        if at("'",prglin) > 0
          delim1 := delim2 := "'"
        elseif at('"',prglin) > 0
          delim1 := delim2 := '"'
        elseif at('[',prglin) > 0
          delim1 := '['
          delim2 := ']'
        else
          alert( 'Unable to analyze ' + prglin )
          break
        endif

        incstart := at(delim1,prglin) + 1
        incstop  := rat(delim2,prglin) - 1
        incfile  := substr( prglin, incstart, incstop-incstart+1 )

        * Determine INCLUDE path and locate file *

        if len(aIPath) = 0 // include path not yet determined

          aadd( aIPath, curdir() ) // no trailing \'s
          incstr := gete('INCLUDE')
          if len(trim(incstr)) > 0
            do while at(";",incstr) > 0
              aadd( aIPath, substr( incstr, 1, at(";",incstr)-1 ) )
              incstr := substr( incstr, at(";",incstr)+1 )
            enddo
            if len(trim(incstr)) > 0
              aadd( aIPath, incstr )
            endif
          endif
          aeval( aIPath, {|e,n| aIPath[n] := e + [\] } ) // add trailing \'s

        endif

        for px = 1 to len(aIPath)
          if file( aIPath[px] + incfile )
            incpath := aIPath[px]
            exit
          endif
        next
        if empty(incpath)
          alert( incfile + ' not found' )
        else

          @ icr,icc say lsi(++inccnt)
          ANALYPRG( incpath + incfile ) // recursive call

        endif

        END SEQUENCE

      endif // incinc

    endif

  endif // in comment / not

  if len(prgstr) > at(chr(K_ENTER),prgstr)
    prgstr := substr( prgstr, at(chr(K_ENTER),prgstr)+1 )
  else
    exit
  endif

enddo // K_ENTER in prgstr

return (NIL)

/*-----------------------------------------------------------*/

Function MCAFunc(mode)
* User function for achoice() in MC()

local retval,;
      lkey := lastkey()

if mode = AC_EXCEPT

  if (lkey >= 65 .and. lkey <= 90) .or. (lkey >= 97 .and. lkey <= 122)  // A-Z,a-z
    retval := AC_GOTO

  elseif lkey = K_ESC
    retval := AC_ABORT

  else
    retval := AC_CONT
  endif

else
  retval := AC_CONT
endif

return (retval)

/*-----------------------------------------------------------*/